# -*- coding: utf-8 -*-
# flake8: noqa
# Copyright 2012-2023 Dr. Jan-Philip Gehrcke. See LICENSE file for details.


__version__ = "1.8.0"


from .gipc import pipe, start_process, GIPCError, GIPCClosed, GIPCLocked
